;;; w32-print.el --- MS Windows printing

;; Copyright (C) 2004, 2005, 2006 by Lennart Borgman

;; Author:     Lennart Borgman <lennart DOT borgman DOT 073 AT student DOT lu DOT se>
;; Created: 2005-08-01
;; Version: 0.73
;; Last-Updated: Thu May 25 01:13:29 2006 (7200 +0200)
;; Keywords: extensions convenience printing
;; URL: http://OurComments.org/Emacs/DL/elisp/w32-print.el
;; Features that might be required by this library:
;;
;;   `button', `cl', `custom', `easymenu', `help-fns', `help-mode',
;;   `htmlize', `htmlize-view', `view', `w32-integ', `widget'.
;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;; Commentary:
;;
;; This file uses htmlize-view.el to show the current buffer in your
;; web browser.  In addition to that it can directly print the buffer,
;; without preview in the web browser, but still using the web
;; browser.  It also adds an entry for quick B/W printing using
;; Notepad.  Printing through the web browser will show the normal
;; dialogs for printing on MS Windows with the normal choices.  It
;; replaces the htmlize-view.el entry on the File menu for this:
;;
;;   - Quick Print (to Windows Printer)
;;
;; The default print entries in Emacs on the File menu can be removed,
;; see the settings `w32-print-menu-show-print' and
;; `w32-print-menu-show-ps-print'.
;;
;; To use this package put in your load-path and do
;;    (require 'w32-print)
;;    (w32-print-add-to-files-menu)
;;
;; If not run on MS Windows this file does nothing except gives a
;; warning message. If you want the (nearly) the same functionality on
;; another OS then just use htmlize-view.el instead.


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;; History:
;;
;; Modified: according to tips from Vinicius Jose Latorre <viniciusjl_@_ig.com.br>
;;   - Changed the check for win32
;;
;;   - Now uses htmlize-view.el
;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; This file is not part of GNU Emacs
;;
;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2, or (at
;; your option) any later version.
;;
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.
;;
;; To find out more about the GNU General Public License you can visit
;; Free Software Foundation's website http://www.fsf.org/.  Or, write
;; to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
;; Floor, Boston, MA 02110-1301, USA.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;; Code:


(if (not (eq system-type 'windows-nt))

    (lwarn '(w32-print) :error "w32-print.el can only be used on MS Windows")

  (require 'w32-integ)
  (require 'htmlize-view)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;; Printing

  (defun w32-print-print-file-notepad(file-name)
    "Print file with notepad."
    (message "w32-shell-execute-with-msg open notepad.exe (concat /p %s" file-name)
    (w32-shell-execute-with-msg "open" "notepad.exe" (concat "/p " file-name)
				))

  (defun w32-print-print-part-notepad(start-point end-point)
    "Print current buffer part with notepad."
    (let ((file
           (make-temp-file
            (expand-file-name "htmlize-view" htmlize-view-temp-dir) nil ".txt")))
      (save-excursion
        (let (deactivate-mark)
          (copy-to-buffer "*w32-print-temp*" start-point end-point)
          (set-buffer "*w32-print-temp*")
          (write-file file nil)
          (w32-print-print-buffer-notepad)
          (if htmlize-view-kill-view-buffers (kill-buffer (current-buffer)))))))


  (defun w32-print-print-region-notepad()
    "Print current region with notepad."
    (interactive)
    (let ((start (mark)) (end (point)))
      (or (<= start end)
	  (setq start (prog1 end (setq end start))))
      (w32-print-print-part-notepad start end)))


  (defun w32-print-print-buffer-notepad()
    "Print current buffer with notepad."
    (interactive)
    (let ((file (buffer-file-name (current-buffer))))
      (if (not file)
	  (progn
	    (save-excursion
	      (widen)
	      (w32-print-print-part-notepad (point-min) (point-max))))
	(if (buffer-modified-p)
	    (let ((use-dialog-box nil))
	      (if (y-or-n-p "File is modified.  Save before printing? ")
		  (save-buffer))))
	(w32-print-print-file-notepad file))))



  (defun w32-print-verb-htmlize(whole-buffer verb)
    "Convert to html and send resulting file to Explorer.
Convert to html, write to a temporary file and send the temporary
html file to Explorer with VERB.

If WHOLE-BUFFER is non-nil then use the whole buffer, otherwise
the region."
    (interactive)
    (let ((html-file (htmlize-buffer-to-tempfile (not whole-buffer))))
      (w32-shell-execute-with-msg verb html-file)))

  (defun w32-print-print-buffer-htmlize()
    "Print buffer as html."
    (interactive)
    (w32-print-verb-htmlize t "print"))

  (defun w32-print-print-region-htmlize()
    "Print region as html."
    (interactive)
    (w32-print-verb-htmlize nil "print"))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;; Menus

  ;;(defun htmlize-view-add-to-files-menu()
  ;;(w32-print-add-to-files-menu)
  (defun w32-print-add-to-files-menu()
    "Add \"Quick Print\" entry to file menu."
    (setq htmlize-view-print-menu (make-sparse-keymap "QP"))
    (setq htmlize-view-print-region-menu (make-sparse-keymap "QPR"))
    ;; Main
    (define-key-after menu-bar-file-menu [htmlize-view-print]
      (list 'menu-item
            "Quick Print (to OS Default Printer)"
            htmlize-view-print-menu
            :visible 'htmlize-view-print-visible)
    'separator-print)
    ;; Main submenu
    (define-key htmlize-view-print-menu [htmlize-view-region]
      (list 'menu-item "Print Region" htmlize-view-print-region-menu
            :help "Print only region"
            :enable 'mark-active))
    (define-key htmlize-view-print-menu [htmlize-view-separator-pre]
      '(menu-item "--"))
    (define-key htmlize-view-print-menu [w32-print-notepad]
      (list 'menu-item "Print with Notepad" 'w32-print-print-buffer-notepad
            :help "Print buffer with Notepad"))
    (define-key htmlize-view-print-menu [htmlize-view-browser-pre]
      '(menu-item "Preview with Web Browser" htmlize-view-buffer
                  :help "Print preview buffer with web browser"
                  :visible t))
    (define-key htmlize-view-print-menu [w32-print-browser]
      (list 'menu-item "Print with Web Browser" 'w32-print-print-buffer-htmlize
            :help "Print buffer with web browser"))
    ;; Submenu region
    (define-key htmlize-view-print-region-menu [w32-print-notepad-region]
      (list 'menu-item "Print with Notepad" 'w32-print-print-region-notepad
            :help "Print Region with Notepad"))
    (define-key htmlize-view-print-region-menu [htmlize-view-browser-region-pre]
      '(menu-item "Preview with Web Browser" htmlize-view-region
                  :help "Print preview region with web browser"))
    (define-key htmlize-view-print-region-menu [w32-print-browser-region]
      (list 'menu-item "Print with Web Browser" 'w32-print-print-region-htmlize
            :help "Print region with web browser"))
    )

;;   (defun w32-print-add-to-files-menu()
;;     "Add \"Quick Print\" entry to file menu."
;;     (let ((print-menu-spec
;; 	   (list
;;             )
;;            ))
;;       (when (featurep 'htmlize-view)
;;         (setq print-menu-spec
;;               (append print-menu-spec
;;                       (list
;;                        ["&Print with Web Browser" (w32-print-print-buffer-htmlize) t
;;                         :help "Print buffer with web browser"
;;                         :key-sequence nil
;;                         ]
;;                        ["Pre&view with Web Browser" (htmlize-view-buffer) t
;;                         :help "Print preview buffer with web browser"
;;                         :key-sequence nil
;;                         ]
;;                        ["Print with &Notepad" (w32-print-print-buffer-notepad)
;;                         :help "Print buffer with Notepad (only black and white)"
;;                         ]
;;                        )
;;                       )))
;;       (let ((submenu
;;              (list "Print &Region"
;;                    :help "Print only region"
;;                    :active 'mark-active
;;                    )))
;;         (when (featurep 'htmlize-view)
;;           (setq submenu
;;                 (append submenu
;;                         (list
;;                          ["&Print with Web Browser" (w32-print-print-region-htmlize) t
;;                           :help "Print region with web browser"
;;                           ]
;;                          ["Pre&view with Web Browser" (htmlize-view-buffer t) t
;;                           :help "Print preview region with web browser"
;;                           ]
;;                          ["Print with &Notepad (B/W)" (w32-print-print-region-notepad) t
;;                           :help "Print Region with Notepad (only black and white)"
;;                           :key-sequence nil
;;                           ]
;;                          ))))
;;         (setq print-menu-spec (append print-menu-spec
;;                                       [["-" nil
;;                                         ;;:visible mark-active
;;                                         ]]
;;                                       (list submenu))))

;;       ;; Remove any entry from htmlize-view.el:
;;       (easy-menu-remove-item nil '("file") "Quick Print (to OS Default Printer)")

;;       (easy-menu-change
;;        ;; This is a dirty check for Emacs older than 21.3
;;        ;;(if (fboundp 'cua-mode) '("file") '("files"))
;;        '("file")
;;        "&Quick Print (to OS Default Printer)"
;;        print-menu-spec "separator-window") ))

  (defun w32-print-set-visible(menu entry visible)
    (let* ((pb (assoc entry menu))
           (hasvis (memq :visible pb))
           (helpcdr (cdr (memq :help pb))))
      (if hasvis
          (progn
            (setq hasvis (cdr hasvis))
            (setcar hasvis visible))
        (unless visible
          (setcdr helpcdr (cons :visible (cons nil (cdr helpcdr))))))))

  (defun w32-print-set-visible-ps-print-entries(visible)
    "Remove default ps-print entries from the file menu."
    ;;(message "ps-print-en %s" visible) (sit-for 4)
    (w32-print-set-visible menu-bar-file-menu 'ps-print-buffer visible)
    (w32-print-set-visible menu-bar-file-menu 'ps-print-region visible)
    (w32-print-set-visible menu-bar-file-menu 'ps-print-buffer-faces visible)
    (w32-print-set-visible menu-bar-file-menu 'ps-print-region-faces visible))

  (defun w32-print-set-visible-print-entries(visible)
    "Remove default print entries from file menu.
They are useless on MS Windows unless the user has added capabilities
for them."
    ;;(message "print-en %s" visible) (sit-for 4)
    (w32-print-set-visible menu-bar-file-menu 'print-buffer visible)
    (w32-print-set-visible menu-bar-file-menu 'print-region visible))

  (defgroup w32-print nil
    "Printing options useful for MS Windows users.
Notice that emacsw32 has support for simple printing (in colors if
you have a color printer) to default MS Windows printers. You find
this in the menus on File - Quick Print.  There is no setup in Emacs
for Quick Print, just print!"
  :tag "W32 Printing setup"
  :group 'emacsw32
  )

  (defcustom w32-print-menu-show-print t
    "If nil remove default print entries from the file menu.
These may be useless unless you do a sometimes rather complicated
setup for printing."
    :tag "Remove default print entries from the file menu"
    :type 'boolean
    :set (lambda (symbol value)
	   (set-default symbol value)
           (w32-print-set-visible-print-entries value))
    :group 'w32-print
    )

  (defcustom w32-print-menu-show-ps-print t
    "If nil remove default ps-print entries from the file menu.
These may be useless unless you do a sometimes rather complicated
setup for printing."
    :tag "Remove default ps-print entries from the file menu"
    :type 'boolean
    :set (lambda (symbol value)
	   (set-default symbol value)
           (w32-print-set-visible-ps-print-entries value))
    :group 'w32-print
    )


  )

(provide 'w32-print)

;;; w32-print.el ends here
