;; Functions to analyse clearcase paths and allow rudimentary
;; conversions between Windows and Unix provided the equivalent view
;; name is defined in both Regions.  E.g. by using the Region
;; Synchroniser on Windows to make Unix views available.

;; Examples:
;;
;;  (clearcase-path-analyse "/net/machine/view/my_vu/vobs/my_vob/directory/path/file@@/main/branch/1")
;; ("/net/machine" "my_vu" "my_vob" "/directory/path" "file" "@@/main/branch/1" "/vobs")
;;
;;  (clearcase-path-analyse "\\\\view\\my_vu\\my_vob\\directory\\path\\file@@\\main\\branch\\1")
;; ("\\\\view" "my_vu" "my_vob" "\\directory\\path" "file" "@@\\main\\branch\\1" "/vobs")
;;
;;  (clearcase-path-convert "\\\\view\\my_vu\\my_vob\\directory\\path\\file@@\\main\\branch\\1" 'unix)
;; "/view/my_vu/vobs/my_vob/directory/path/file@@/main/branch/1"
;;
;;  (clearcase-path-convert "/net/machine/view/my_vu/vobs/my_vob/directory/path/file@@/main/branch/1" 'windows
;;   'view-drive "\\\\view")
;; "\\\\view\\my_vu\\my_vob\\directory\\path\\file@@\\main\\branch\\1"
;;
;;  (clearcase-path-convert "/net/machine/view/my_vu/vobs/my_vob/directory/path/file@@/main/branch/1" 'windows)
;; "m:\\my_vu\\my_vob\\directory\\path\\file@@\\main\\branch\\1"

;;(require 'clearcase)

(defvar clearcase-path-unix-format-re
    "^\\([/\\]net[/\\][^/\\]+\\)?[/\\]view[/\\]\\([^/\\]+\\)%s/\\([^/\\]+\\)\\(.*\\)"
  "The Regular expression string to match Unix paths.
It is printf format to allow setting of the vob-mount path to vobs.
It also allows the path to be in an automount form i.e. under /net/")

(defvar clearcase-path-windows-format-re
    "^\\(%s\\|[/\\][/\\]view\\)[/\\]\\([^/\\]+\\)[/\\]\\([^/\\]+\\)\\(.*\\)"
    "The Regular expression string to match Windows paths.
It is printf format to allow setting of the MVFS drive letter used for the \\view mount.
It also allows the path to be in UNC form ie. using \\\\view\\...")

(defvar clearcase-path-cleartool-unix-format-re
    "^%s/\\([^/\\]+\\)\\(.*\\)"
  "The Regular expression string to match Unix paths in a cleartool shell.
It is printf format to allow setting of the vob-mount path to vobs.
It is assumed that the CLEARCASE_ROOT environment variable contains the /view/ path.")

(defvar clearcase-path-view-drive (if (boundp 'clearcase-view-drive)
				      clearcase-view-drive
				    "m:")
  "The MVFS drive letter used for the \\\\view mount.
It is set via the clearcase.el variable `clearcase-view-drive' if defined
or the normal Clearcase drive letter default which is m:.")

(defvar clearcase-path-xn-sfx (if (boundp 'clearcase-vxpath-glue)
				  clearcase-vxpath-glue
				"@@")
  "The string used to construct version-extended pathnames.
It is set via the clearcase.el variable `clearcase-vxpath-glue' if defined
otherwise defaults to @@.")

(defun clearcase-path-analyse (path &rest plist)
  "Deconstruct a path into its constituent parts.
Returns a list: (Root, View, Vob, dir, file, extended path, unix-vob-mount)
Optional property list:
    unix-vob-mount  The mount point for vobs on Unix. Default: /vobs"

  (let ((root "") (view "") (vob "") (dir "") (file "") (xpath "")
	(unix-vob-mount (or (plist-get plist 'unix-vob-mount) "/vobs"))
	rest
	dir-file branch)
    (save-match-data
      (cond ((string-match (format clearcase-path-unix-format-re unix-vob-mount) path)
	     (setq root (or (match-string 1 path) "")
		   view (match-string 2 path)
		   vob  (match-string 3 path)
		   rest (or (match-string 4 path) "/")))
	    ((string-match (format clearcase-path-windows-format-re clearcase-path-view-drive) path)
	     (setq root (match-string 1 path)
		   view (match-string 2 path)
		   vob  (match-string 3 path)
		   rest (or (match-string 4 path) "\\")))
	    ((and (getenv "CLEARCASE_ROOT")
		  (string-match (format clearcase-path-cleartool-unix-format-re unix-vob-mount) path))
	     (setq root ""
		   vob  (match-string 1 path)
		   rest (or (match-string 2 path) "/"))
	     (string-match "^[/\\]view[/\\]\\([^/\\]+\\)" (getenv "CLEARCASE_ROOT"))
	     (setq view (match-string 1 (getenv "CLEARCASE_ROOT"))))
	    (t
	     nil))

      (if rest
	  ;;We have found a Clearcase path
	  (progn
	    (setq rest (split-string rest clearcase-path-xn-sfx))
	    (setq dir-file (car rest))
	    (string-match "\\(.*\\)[/\\]\\([^/\\]+\\)$" dir-file) ;;FATAL error if path ends with /
	    (setq dir    (match-string 1 dir-file)
		  file   (match-string 2 dir-file)
		  branch (cadr rest))


	    (if branch
		(setq xpath (concat clearcase-path-xn-sfx branch))))))
    (list root view vob dir file xpath unix-vob-mount)))

(defun clearcase-path-convert (path &optional convert &rest plist)
  "Convert a fully qualified clearcase path between Windows and Unix formats.
To convert to Unix    set convert to 'unix    or be on Unix    machine.
To convert to Windows set convert to 'windows or be on Windows machine.

Unix:    [/net/machine]/view/view_name/unix_vob_mount/vob_name
Windows: \\\\view\\view_name\\vob_name
         VIEWDRIVE:\\view_name\\vob_name (Where VIEWDRIVE is the MVFS drive)

Optional property list:
    unix-vob-mount  The mount point for vobs on Unix.
    view-drive      The Root drive or UNC path to use for Windows."

  (let ((convert (or convert (eq system-type 'windows-nt)))
	(components (apply 'clearcase-path-analyse path plist)))
    (apply 'clearcase-path-rebuild-from-components 
	   (if (eq convert 'unix) nil convert)
	   components
	   plist)))

(defun clearcase-path-rebuild-from-components (windows-q components &rest plist)
  "Rebuild path from component list like that returned by `clearcase-path-analyse'.
The first argument determines whether to return a Windows or Unix path.
May supply unix-vob-mount or view-drive to help get the path correct in the more obscure cases.

Optional property list:
    unix-vob-mount  The mount point for vobs on Unix.
    view-drive      The Root drive or UNC path to use for Windows."

  (let (path
	(root  (nth 0 components))
	(view  (nth 1 components))
	(vob   (nth 2 components))
	(dir   (nth 3 components))
	(file  (nth 4 components))
	(xpath (nth 5 components))
	(unix-vob-mount (or (plist-get plist 'unix-vob-mount) (nth 6 components)))
	(view-drive     (or (plist-get plist 'view-drive)     "\\\\view")))

    (save-match-data
      (cond (windows-q
	     (setq root (if (and (not (equal root "")) ;Ignore Unix automount path
				 (not (string-match "/net/" root)))
			    root
			  view-drive)
		   path (concat root "\\" view "\\" vob dir "\\" file xpath))
	     (subst-char-in-string ?/ ?\\ path t))
	    (t
	     (setq root (if (string-match "/net/" root) ;Only allow root to be an automount path.
			    root
			  "")
		   path (concat root "/view/" view unix-vob-mount "/" vob dir "/" file xpath))
	     (subst-char-in-string ?\\ ?/ path t))))
    path))

(provide 'clearcase-path)
