;;; vc-patches.el: contains some bug fixes to the vc code to enable it to run on Windows

;; The vc-version-other-window [C-x v ~] function (which is also
;; used inside ediff-revision [C-x v =]) uses a /bin/sh shell
;; script to check out a file to stdout (using -p) and redirect
;; output to a designated file.
;;
;; This patch (for Windows NT) recognizes that case and fixes it
;; to work with the regular CMD command shell.
;;
;; vc-do-command is the best place to trap and fix this.  We use the
;; advice package to achieve the patch.
;;
;; For the case we want to trap, the arguments are as follows:
;; * arg 2:  "/bin/sh"
;;   arg 3:  the file under RCS
;;   arg 4:  'MASTER
;; * arg 5:  "-c"
;; * arg 6:  the /bin/sh script
;; * arg 7:  ""
;;   arg 8:  file to redirect to
;;   arg 9+: options such as "-l" "-p[rev]", etc.
;;
;; We actually look for a suitable match for all asterisked arguments
;; before patching them.  We patch the arguments so the invocation ends
;; up instead being of the form:
;;     cmd /c co >outfile args
;;
;; Since I believe that this bug has been fixed for Emacs 21 it is not applied
(if (and (eq system-type 'windows-nt)
	 (<= emacs-major-version 20))
    (defadvice vc-do-command (before patch-vc-do-command-for-nt-rcs activate)
      "Patch vc-do-command to work under NT, so vc-version-other-window and ediff-revision can work."
      ;; First, let's check for the case we want to patch:
      (if (and (string= (ad-get-arg 2) "/bin/sh")
	       (string= (ad-get-arg 5) "-c")
	       (string-match (concat "^if \\[ x\"\\$1\" = x \\]; then shift; fi;[ \t]+"
				     "umask 333; exec >\"\$1\" || exit;[ \t]+shift; "
				     "umask 0; exec co \"\\$@\"") (ad-get-arg 6))
	       (string= (ad-get-arg 7) ""))
	  (progn (ad-set-arg 2 "cmd") ;; Invoke "cmd" instead of "/bin/sh"
		 (ad-set-arg 5 "/c")  ;; Convert "-c" to "/c"
		 (ad-set-arg 6 nil)   ;; Zap the /bin/sh script
		 (ad-set-arg 7 "co")  ;; Invoke the "co" command
		 (ad-set-arg 8 (concat ">" (ad-get-arg 8))) ;; Redirect to outfile
		 ))))
;; Same for CVS
(if (eq system-type 'windows-nt)
    (defadvice vc-do-command (before patch-vc-do-command-for-nt-cvs activate)
      "Patch vc-do-command to work under NT, so vc-version-other-window and ediff-revision can work."
      ;; First, let's check for the case we want to patch:
      (if (and (string= (ad-get-arg 2) "/bin/sh")
	       (string= (ad-get-arg 5) "-c")
	       (string-match "exec >\"$1\" || exit; shift; exec cvs update \"$@\"" (ad-get-arg 6))
	       (string= (ad-get-arg 7) ""))
	  (progn (ad-set-arg 2 "cmd") ;; Invoke "cmd" instead of "/bin/sh"
		 (ad-set-arg 5 "/c")  ;; Convert "-c" to "/c"
		 (ad-set-arg 6 nil)   ;; Zap the /bin/sh script
		 (ad-set-arg 7 "cvs update")  ;; Invoke the "cvs" command
		 (ad-set-arg 8 (concat ">" (ad-get-arg 8))) ;; Redirect to outfile
		 ))))


(if (eq system-type 'windows-nt)
    (defadvice start-process (before patch-start-process-for-nt activate)
      "Patch start-process for cvs-update to work under NT."
      ;; First, let's check for the case we want to patch:
      (if (and (string= (ad-get-arg 0) "cvs")
	       (string= (ad-get-arg 2) "/bin/sh")
	       (string= (ad-get-arg 3) "-c")
	       ;(string-match "" (ad-get-arg 6))
	       )
	  (progn (ad-set-arg 2 "cmd") ;; Invoke "cmd" instead of "/bin/sh"
		 (ad-set-arg 3 "/c")  ;; Convert "-c" to "/c"
		 ))))

(provide 'vc-patches)
